import argparse
import sys
from os import path, mkdir, chdir


def run():
    parser = argparse.ArgumentParser(prog='pastry init')
    parser.add_argument('directory', nargs='?', type=str,
                        help='Path to directory to create (or default to current directory)')
    args = parser.parse_args(sys.argv[2:])  # type: argparse.Namespace

    directory = args.directory if args.directory is not None else '.'

    try:
        mkdir(directory)
    except FileExistsError:
        pass

    chdir(directory)



import pandas as pd
from sklearn.metrics import f1_score, make_scorer
from sklearn.model_selection import StratifiedKFold

from sklearn.tree import DecisionTreeClassifier

from pypastry.experiment import Experiment


def get_experiment():
    dataset = pd.DataFrame({
        'feature': [1, 0, 1, 1, 0, 0, 1, 1, 0, 1],
        'class': [True, False, True, True, False, False, True, True, False, False],
    })
    predictor = DecisionTreeClassifier()
    cross_validator = StratifiedKFold(n_splits=5)
    scorer = make_scorer(f1_score)
    label_column = 'class'
    return Experiment(dataset, label_column, predictor, cross_validator, scorer)



import json
from datetime import datetime
from types import ModuleType
from typing import Any, Dict, Tuple, List

import numpy as np
import pandas as pd
from git import Repo
from joblib import Parallel, delayed
from pandas import Series
from sklearn.base import BaseEstimator, is_classifier, clone
from sklearn.metrics._scorer import _BaseScorer
from sklearn.model_selection import check_cv, PredefinedSplit

from pypastry import display
from pypastry.experiment import Experiment
from pypastry.experiment.hasher import get_dataset_hash
from pypastry.experiment.results import ResultsRepo
from pypastry.paths import REPO_PATH, RESULTS_PATH

MAX_PARAMETER_VALUE_LENGTH = 500


class DirtyRepoError(Exception):
    def __init__(self, message):
        super().__init__(message)


class ExperimentRunner:
    def __init__(self, git_repo: Repo, results_repo: ResultsRepo, results_display: ModuleType):
        self.git_repo = git_repo
        self.results_repo = results_repo
        self.results_display = results_display

    def run_experiment(
        self,
        experiment: Experiment,
        message: str = "",
        force: bool = False,
        limit: int = None,
    ):

        print("Got dataset with {} rows".format(len(experiment.dataset)))
        if force or not self.git_repo.is_dirty():
            print("Running evaluation")
            estimators = self._run_evaluation(experiment, message)
            results = self.results_repo.get_results(self.git_repo)
            self.results_display.cache_display(results)
        else:
            raise DirtyRepoError("There are untracked/unstaged/staged changes in git repo, force flag was not given. "
                                 "Please commit your changes or provide force flag - note that in this case "
                                 "saved commit hash in your result file will not correspond to the actual code!")
        self.results_display.print_cache_file(limit)
        return estimators

    def _run_evaluation(self, experiment: Experiment, message: str):
        run_info, estimators = evaluate_predictor(experiment)
        dataset_hash = get_dataset_hash(experiment.dataset, experiment.test_set)
        dataset_info = {
            'hash': dataset_hash,
            'columns': experiment.dataset.columns.tolist(),
        }
        git_info = {
            "git_hash_msg": ("dirty_" if self.git_repo.is_dirty() else "") + self.git_repo.head.object.hexsha[:8],
            "git_summary_msg": message,
        }
        _ = self.results_repo.save_results(run_info, dataset_info, git_info=git_info)
        return estimators


def evaluate_predictor(experiment: Experiment) -> Dict[str, Tuple[Any, List[BaseEstimator]]]:
    start = datetime.utcnow()
    scores, estimators = _get_scores_and_estimators(experiment)
    end = datetime.utcnow()

    additional = experiment.additional_info
    additional_info = [additional(estimator) if additional is not None else None
                       for estimator in estimators]

    if experiment.group_column is not None:
        for group, score_values in scores:
            score_values[experiment.group_column] = group

    values = [x[1] for x in scores]

    scores_array = pd.DataFrame(values)

    mean_score = scores_array.mean().to_dict()
    sem_score = scores_array.sem().to_dict()
    results = {'test_score': mean_score, 'test_score_sem': sem_score}

    model_info = get_model_info(experiment.predictor)

    run_info = {
        'run_start': str(start),
        'run_end': str(end),
        'run_seconds': (end - start).total_seconds(),
        'results': results,
        'results_detail': scores_array.to_dict('list'),
        'model_info': model_info,
        'additional_info': additional_info,
    }

    return run_info, estimators


def get_model_info(model: BaseEstimator):
    all_info = model.get_params()
    info = {key: value for key, value in all_info.items()
            if len(json.dumps(value, default=str)) < MAX_PARAMETER_VALUE_LENGTH}
    info['type'] = type(model).__name__
    return info


def _get_scores_and_estimators(experiment: Experiment) -> Tuple[List[float], List[Any]]:
    if experiment.test_set is not None:
        assert experiment.cross_validator is None, "Cannot use a cross validator with train test split"
        dataset = pd.concat([experiment.dataset, experiment.test_set])
        split = np.array([-1] * len(experiment.dataset) + [1] * len(experiment.test_set))
        cross_validator = PredefinedSplit(split)
    else:
        dataset = experiment.dataset
        cross_validator = experiment.cross_validator

    X = dataset.drop(columns=[experiment.label_column])
    y = dataset[experiment.label_column]
    if experiment.group_column is None:
        if experiment.average_scores_on_instances:
            groups = Series(range(len(X)), index=X.index)
        else:
            groups = None
    else:
        groups = X[experiment.group_column]
        X = X.drop(columns=[experiment.group_column])

    cv = check_cv(cross_validator, y, classifier=is_classifier(experiment.predictor))
    train_test = cv.split(X, y, groups)

    # We clone the estimator to make sure that all the folds are
    # independent, and that it is pickle-able.
    parallel = Parallel(n_jobs=None, verbose=False,
                        pre_dispatch='2*n_jobs')
    scores_and_estimators = parallel(
        delayed(_fit_and_predict)(
            clone(experiment.predictor), X, y, train, test, groups, experiment.scorer)
        for train, test in train_test)
    scores_lists, estimators = zip(*scores_and_estimators)
    scores = [score for score_list in scores_lists for score in score_list]
    return scores, estimators


def _fit_and_predict(estimator: BaseEstimator, X, y, train, test, groups, scorer):
    if groups is not None:
        scores = _fit_and_predict_groups(X, estimator, groups, scorer, test, train, y)
    else:
        scores = _fit_and_predict_simple(X, estimator, scorer, test, train, y)
    return scores, estimator


def _fit_and_predict_simple(X, estimator, scorers, test, train, y):
    X_train = X.iloc[train]
    y_train = y.iloc[train]
    estimator.fit(X_train, y_train)
    X_test = X.iloc[test]
    y_test = y.iloc[test]
    score = _score(scorers, estimator, X_test, y_test)
    return [(None, score)]


def _fit_and_predict_groups(X, estimator, groups, scorers, test, train, y):
    X_train = X.iloc[train]
    y_train = y.iloc[train]
    estimator.fit(X_train, y_train)
    X_test = X.iloc[test]
    y_test = y.iloc[test]
    groups_test = groups.iloc[test]
    test_df = pd.DataFrame(X_test)
    test_df['y'] = y_test
    test_df['groups'] = groups_test
    test_groups = test_df.groupby('groups')
    scores = []
    for key, group in test_groups:
        X_group = group[X.columns]
        score = _score(scorers, estimator, X_group, group['y'])
        scores.append((key, score))
    return scores


def _score(scorers: List[_BaseScorer], estimator, X_test, y_test):
    scores = {}
    for scorer in scorers:
        score = scorer(estimator, X_test, y_test)
        score_name = scorer._score_func.__name__
        sign = scorer._sign
        score_ignoring_sign = score*sign
        scores[score_name] = score_ignoring_sign
    return scores


def run_experiment(experiment, message="", force=False):
    git_repo = Repo(REPO_PATH, search_parent_directories=True)  # type: pypastry.experiment.Experiment
    results_repo = ResultsRepo(RESULTS_PATH)  # type: pypastry.experiment.results.ResultsRepo
    runner = ExperimentRunner(git_repo, results_repo, display)  # type:
    # pypastry.experiment.evaluation.ExperimentRunner
    return runner.run_experiment(experiment, message, force)



import argparse
import sys

from pypastry.experiment.evaluation import run_experiment


def run():
    parser = argparse.ArgumentParser(prog='pastry run')
    parser.add_argument('-m', '--message', default="", type=str, help='Summary message about the experiment.')
    parser.add_argument('-f', '--force', action='store_true', help='Force a re-run of the experiment')

    args = parser.parse_args(sys.argv[2:])

    sys.path.append('.')
    import pie
    experiment = pie.get_experiment()
    force = args.force
    message = args.message

    run_experiment(experiment, message, force)



from unittest.mock import Mock

from pandas import DataFrame
from sklearn.dummy import DummyClassifier
from sklearn.metrics import accuracy_score, make_scorer, precision_score
from sklearn.model_selection import StratifiedShuffleSplit, GroupShuffleSplit
from sklearn.tree import DecisionTreeClassifier

from pypastry.experiment import Experiment
from pypastry.experiment.evaluation import ExperimentRunner, evaluate_predictor


def test_simple_evaluation(dirty=False, force=False):
    # Given
    dataset = DataFrame({
        'a': [1, 1, 0, 0],
        'b': [1, 1, 0, 0],
    })

    cross_validation = StratifiedShuffleSplit(n_splits=1, test_size=0.5)
    predictor = DecisionTreeClassifier()
    scorer = make_scorer(accuracy_score)
    experiment = Experiment(dataset, 'b', predictor, cross_validation, scorer)

    git_mock = Mock()
    git_mock.is_dirty.return_value = dirty
    results_repo_mock = Mock()
    new_results_files = ['results/abc.json']
    results_repo_mock.save_results.return_value = new_results_files
    results_display_mock = Mock()
    runner = ExperimentRunner(git_mock, results_repo_mock, results_display_mock)

    # When
    runner.run_experiment(experiment, "", force)

    # Then
    call_args_list = results_repo_mock.save_results.call_args_list
    assert 1 == len(call_args_list)
    run_info, dataset_info = call_args_list[0][0]
    print("Run info", run_info)

    results = run_info['results']
    assert {'accuracy_score': 1.0} == results['test_score']
    assert ['a', 'b'] == dataset_info['columns']

    # TODO: check the hash. Need to find a way to make this consistent between python versions etc.
    # assert '28ea628a50a47c726a9b0ec437c88fc4742d81fd' == dataset_info['hash']

    assert 1 == len(results_display_mock.cache_display.call_args_list)
    print(results_display_mock.cache_display.call_args[0])
    assert len(results_display_mock.cache_display.call_args[0]) > 0
    assert 1 == len(results_display_mock.print_cache_file.call_args_list)


def test_grouped_evaluation():
    label = [i % 2 for i in range(100)]
    dataset = DataFrame({
        'a': label,
        'b': label,
        'g': [i // 2 for i in range(100)]
    })

    cross_validation = GroupShuffleSplit(n_splits=1, test_size=0.5)
    predictor = DummyClassifier(strategy='constant', constant=1)
    scorer = make_scorer(accuracy_score)
    experiment = Experiment(dataset, 'b', predictor, cross_validation, scorer, group_column='g')

    git_mock = Mock()
    git_mock.is_dirty.return_value = False
    results_repo_mock = Mock()
    results_display_mock = Mock()
    runner = ExperimentRunner(git_mock, results_repo_mock, results_display_mock)

    runner.run_experiment(experiment, "", False)

    assert 1 == len(results_repo_mock.save_results.call_args_list)
    run_info, dataset_info = results_repo_mock.save_results.call_args[0]

    print("Run infos", run_info)

    expected_results = {'test_score': {'accuracy_score': 0.5}, 'test_score_sem': {'accuracy_score': 0.0}}
    assert expected_results == run_info['results']


def test_multiple_scorers():
    # Given
    dataset = DataFrame({
        'a': [1, 1, 0, 0],
        'b': [1, 1, 0, 0],
    })

    cross_validation = StratifiedShuffleSplit(n_splits=2, test_size=0.5)
    predictor = DummyClassifier(strategy='constant', constant=1)
    scorer = [make_scorer(accuracy_score), make_scorer(precision_score)]
    experiment = Experiment(dataset, 'b', predictor, cross_validation, scorer)

    run_info, _ = evaluate_predictor(experiment)
    results = run_info['results']
    print("Results", results)

    expected_results = {
        'test_score': {'accuracy_score': 0.5, 'precision_score': 0.5},
        'test_score_sem': {'accuracy_score': 0.0, 'precision_score': 0.0},
    }

    assert expected_results == results



"""
Handle displaying results.

This code needs to display results really fast. That's why it has some odd things:
 - Inline imports
 - Lack of typing for the results_repo to avoid unnecessary imports

This is because I don't like having to wait a second for things to be imported.
I want my pastry now!
"""
import os
from typing import Any, Dict, List, Iterator, TYPE_CHECKING

from pandas import Series

from pypastry.paths import DISPLAY_PATH, DISPLAY_DIR
if TYPE_CHECKING:
    import pypastry


def cache_display(results_from_repo: Iterator['pypastry.experiment.results.Result']) -> None:
    results_dataframe = _get_results_dataframe(results_from_repo)
    display = repr(results_dataframe)

    try:
        os.mkdir(DISPLAY_DIR)
    except FileExistsError:
        pass

    with open(DISPLAY_PATH, 'w') as output_file:
        output_file.write(display)


def _get_results_dataframe(results_from_repo: Iterator['pypastry.experiment.results.Result']) -> 'DataFrame':
    from pandas import DataFrame, set_option
    set_option('display.max_rows', None)
    set_option('display.max_columns', None)
    set_option('display.width', None)
    set_option('display.max_colwidth', -1)
    results = []
    for repo_result in results_from_repo:
        data = repo_result.data
        result = {
            'Git hash': data["git_hash"] if "git_hash" in data else "Unavailable",
            'Git summary': data["git_summary"] if "git_summary" in data else "Unavailable",
            'Dataset hash': data['dataset']['hash'][:8],
            'Run start': data['run_start'][:19],
            'Model': data['model_info']['type'],
            'Duration (s)': "{:.2f}".format(data['run_seconds']),
        }

        try:
            scores = DataFrame(data['results'])
            for row in scores.itertuples():
                result[row.Index] = "{:.3f} ± {:.3f}".format(row.test_score, row.test_score_sem)
        except ValueError:
            result['Score'] = "{:.3f} ± {:.3f}".format(data['results']['test_score'],
                                                       data['results']['test_score_sem'])

        results.append(result)
    results_dataframe = DataFrame(results)
    return results_dataframe.sort_values(by='Run start').reset_index()


def print_cache_file(limit=False):
    with open(DISPLAY_PATH) as display_file:
        print(display_file.read())
        # read_lines = display_file.read()
        # read_list = read_lines.split("\n")
        # if limit:
        #     limit = min(limit, len(read_list)-3)
        #     # to avoid printing more than there is
        # else:
        #     limit = len(read_list)-3
        # print(read_list[0])
        # print("\n".join(read_list[-(2+limit):-2]))
        # print(read_list[-1])




import json
import os
import glob
from tempfile import NamedTemporaryFile
from typing import Dict, Any, List, NamedTuple


Result = NamedTuple('Result', [('data', Dict[str, Any]), ("dirty", bool)])


class ResultsRepo:
    def __init__(self, results_path: str):
        self.results_path = results_path

    def save_results(self, run_info: Dict[str, Any], dataset_info: Dict[str, Any], git_info: Dict[str, str]) -> List[str]:
        try:
            os.mkdir(self.results_path)
        except FileExistsError:
            pass
        new_filenames = []
        run_info['dataset'] = dataset_info
        run_info['git_hash'] = git_info["git_hash_msg"]
        run_info['git_summary'] = git_info["git_summary_msg"]
        with NamedTemporaryFile(mode='w', prefix='result-', suffix='.json',
                                dir=self.results_path, delete=False) as output_file:
            json.dump(run_info, output_file, indent=4, default=str)
            output_file.flush()
        new_filenames.append(output_file.name)
        return new_filenames

    def get_results(self, git_repo):
        for path in glob.glob(os.path.join(self.results_path, "*.json")):
            with open(str(path), "r") as results_file:
                result_json = json.load(results_file)
            yield Result(result_json, git_repo.is_dirty())



import argparse

from pypastry.display import print_cache_file, cache_display, _get_results_dataframe
from pypastry.paths import RESULTS_PATH


def run():
    parser = argparse.ArgumentParser(prog='pastry print')
    parser.add_argument('-l', '--limit', type=int, default=None, help='Limit lines to print')
    parser.add_argument('-e', '--export', type=str, required=False, help='File to output the results in CSV format')

    args = parser.parse_args(sys.argv[2:])
    if args.export is not None:
        results = get_results()
        results_dataframe = _get_results_dataframe(results)
        results_dataframe.to_csv(args.export)
        return
    try:
        print_cache_file(args.limit)
    except FileNotFoundError:
        results = get_results()
        cache_display(results)
        print_cache_file(args.limit)


def get_results():
    from pypastry.experiment.results import ResultsRepo
    from git import Repo
    results_repo = ResultsRepo(RESULTS_PATH)
    results = results_repo.get_results(Repo('.', search_parent_directories=True))
    return results



import sys

from setuptools import setup, find_packages

#check to make sure the python version is compatible 
if sys.version_info < (3, 6):
    sys.exit('Sorry, PyPastry requires Python version 3.6 or greater')

# Reading in the ReadMe file as the doc file
with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pypastry',
    version='0.2.0',
    description='PyPastry machine learning experimentation framework',
    author='Daoud Clarke',
    url='https://github.com/datapastry/pypastry',
    scripts=['pastry'],
    install_requires=['tomlkit', 'pandas', 'scikit-learn', 'pyarrow', 'gitpython'],
    #To find the packages 
    packages=find_packages(),
    #To read in data file modules 
    py_modules=['data/pie'],
    # commands that can be run in a console in the commands folder
    entry_points={
        'console_scripts': [
            'init = pypastry.commands.init:run',
            'print = pypastry.commands.print_:run',
            'run = pypastry.commands.run:run'
            ]},
    package_data={

         '' : ['data/*.gitignore'],
        # And include any *.gitignore files found in the 'data' package, too:
        'data': ['*.gitignore'],

    },
    long_description=long_description,
    long_description_content_type='text/markdown',
    # Make the setup file aware of the Manifest file
    include_package_data=True,
    #Minimum requirement of python, licesnse, and operating system. 
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"],
    python_requires='>=3.5',
    )



from typing import Any, Callable, Union, Iterable

from pandas import DataFrame
from sklearn.base import BaseEstimator
from sklearn.metrics import accuracy_score, make_scorer
from sklearn.metrics._scorer import _BaseScorer as BaseScorer


class Experiment:
    def __init__(self, dataset: DataFrame, label_column: str, predictor: BaseEstimator,
                 cross_validator: Any = None, scorer: Union[BaseScorer, Iterable[BaseScorer]] = None,
                 group_column: str=None, test_set: DataFrame = None, average_scores_on_instances: bool = False,
                 additional_info: Callable[[BaseEstimator], Any] = None):
        if (test_set is not None) == (cross_validator is not None):
            raise ValueError("You must specify either a cross validator or a test set (and not both)")

        if average_scores_on_instances and group_column is not None:
            raise ValueError("You can only average on instances when not grouping instances")

        if scorer is None:
            scorer = [make_scorer(accuracy_score)]

        if not isinstance(scorer, Iterable):
            if not isinstance(scorer, BaseScorer):
                raise ValueError("Scorer must be created using make_scorer()")

            scorer = [scorer]

        self.dataset = dataset
        self.label_column = label_column
        self.predictor = predictor
        self.cross_validator = cross_validator
        self.scorer = scorer
        self.group_column = group_column
        self.test_set = test_set
        self.average_scores_on_instances = average_scores_on_instances
        self.additional_info = additional_info



import json

from pypastry.display import _get_results_dataframe
from pypastry.experiment.results import Result


RESULTS = {
    "run_start": "2020-01-31 09:19:44.261276",
    "run_end": "2020-01-31 09:19:48.743458",
    "run_seconds": 4.482182,
    "results": {
        "test_score": {
            "mean_relative_error": 0.5,
            "mean_absolute_error": 100,
            "mean_squared_error": 1000
        },
        "test_score_sem": {
            "mean_relative_error": 0.01,
            "mean_absolute_error": 1.0,
            "mean_squared_error": 10.0
        }
    },
    "model_info": {
        "n_neighbors": 10,
        "type": "KNearestNeighbor"
    },
    "additional_info": [
        None
    ],
    "dataset": {
        "hash": "998c9dea0afb12d91a8c67f256f80b0a603dd59b",
        "columns": [
            "input",
            "output"
        ]
    },
}


def test_get_display():
    result = Result(RESULTS, 'abc123')
    results_dataframe = _get_results_dataframe([result])
    row = results_dataframe.iloc[0].to_dict()

    expected = {
        'Git hash': 'abc123',
        'Dataset hash': '998c9dea',
        'Run start': '2020-01-31 09:19:44',
        'Model': 'KNearestNeighbor',
        'Duration (s)': '4.48',
        'mean_relative_error': '0.500 ± 0.010',
        'mean_absolute_error': '100.000 ± 1.000',
        'mean_squared_error': '1000.000 ± 10.000',
    }
    assert expected == row
